# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import face_recognition
import numpy as np

def id_detect(image, known_face_encodings, known_face_ids):
    redim = 4

    # Redimensionamos la imagen a 1/4 del tamaño para realizar un reconocimiento más rápido
    small_frame = cv2.resize(image, (0, 0), fx=1/redim, fy=1/redim)

    # Convierte el espacio BGR (opencv) a RGB (face_recognition)
    rgb_small_frame = cv2.cvtColor(small_frame, cv2.COLOR_BGR2RGB)

    # Impedimos que la imagen sea editable (mejora rendimiento)
    rgb_small_frame.flags.writeable = False

    # Localiza todas las caras y sus codificaciones en la imagen
    face_locations = face_recognition.face_locations(rgb_small_frame, model="hog")
    face_encodings = face_recognition.face_encodings(rgb_small_frame, face_locations, model="small")

    id = "desconocido"

    for face_encoding in face_encodings:
        id = "desconocido"  # Desconocido
        # Compara la codificación de una cara de la imagen con las codificaciones de nuestras caras almacenadas
        matches = face_recognition.compare_faces(known_face_encodings, face_encoding)

        # Buscamos la cara conocida con la menor distancia a la nueva cara
        face_distances = face_recognition.face_distance(known_face_encodings, face_encoding)
        best_match_index = np.argmin(face_distances)

        if matches[best_match_index] and face_distances[0] < 1:
            id = known_face_ids[best_match_index]
    return id

def emotion_detect(image, emotion_recognizer):
    image = cv2.resize(image, (150, int(150*1.25)), interpolation=cv2.INTER_CUBIC)
    image = cv2.normalize(image, None, alpha=0, beta=200, norm_type=cv2.NORM_MINMAX)

    result = emotion_recognizer.predict(image)

    return result