# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import mediapipe as mp
import numpy as np
import face_recognition
from deepface import DeepFace
import keras
import os
import sys
import time
import matplotlib.pyplot as plt
from pynput import keyboard

import detect_pose as dp
import face as f2
import show_results as sr

global pulse

def on_press(key):
    global pulse
    pulse = True

if __name__ == '__main__':
    pulse = None
    listener = keyboard.Listener(on_press=on_press)
    listener.start()

    # ¿Pintar imagen con información?
    printed_id = True
    printed_nose = True
    printed_emotion = True
    printed_body = True
    printed_engagement = True
    printed_array = True

    # Objetos detectar cara
    mp_face_detection = mp.solutions.face_detection
    face_detector = mp_face_detection.FaceDetection(min_detection_confidence=0.6)

    # Objetos identificar cara
    known_face_encodings = []
    known_face_ids = []
    last_nose = {}

    # Carga de imagenes de reconocimiento facial
    for id_image in os.listdir("facial_recognition/known_images"):
        image = face_recognition.load_image_file("facial_recognition/known_images/" + id_image)
        known_face_encodings.append(face_recognition.face_encodings(image)[0])
        known_face_ids.append(os.path.splitext(id_image)[0])

    # Objetos detectar emociones
    backends = ['opencv', 'ssd', 'dlib', 'mtcnn', 'retinaface', 'mediapipe']
    emotion_dict = {"angry": "enfado",
                    "disgust": "disgusto",
                    "fear": "miedo",
                    "happy": "feliz",
                    "sad": "Triste",
                    "surprise": "Sorpresa",
                    "neutral": "neutral"}

    # Objetos cara
    mp_face_mesh = mp.solutions.face_mesh
    face_mesh = mp_face_mesh.FaceMesh(max_num_faces=3, min_detection_confidence=0.5, min_tracking_confidence=0.5)

    # Objetos torso
    mp_pose = mp.solutions.pose
    pose = mp_pose.Pose(min_detection_confidence=0.5, min_tracking_confidence=0.5)

    array_memory = []           # Almacena los datos registrados de toda la interacción
    desc = []                   # Almacena caras desconocidas y cada cierto tiempo intenta reconocerlas
    atencion_dict = {}          # Registro de las últimas evaluaciones de la interacción con SVM
    atencion_dict_keras = {}    # Registro de las últimas evaluaciones de la interacción con NN (keras)

    margin = 20                 # Márgenes del recorte de las caras detectadas

    # Clases de niveles de interacción
    name_array_video = ["muy atento", "atento", "poco atento", "despistado", "muy despistado"]

    # Carga de modelo pre-entrenado SVM
    try:
        svm = cv2.ml.SVM_load('model_trainning/recursos/result_model_SVM.xml')
        svm.setType(cv2.ml.SVM_C_SVC)
        svm.setKernel(cv2.ml.SVM_RBF)
    except:
        print("Modelo pre-entrenado SVM no localizado.")
        sys.exit(1)

    # Carga de modelo pre-entrenado Keras
    try:
        keras_model = keras.models.load_model('model_trainning/recursos/result_model_NN_Keras.h5', compile=True)
    except:
        print("Modelo pre-entrenado Keras de redes neuronales no localizado.")
        sys.exit(2)

    # Capturar recurso de la cámara
    cap = cv2.VideoCapture(0)

    while cap.isOpened():
        start_time = time.time()
        #############################
        success, image = cap.read()

        image = cv2.flip(image, 1)

        results = face_detector.process(image)

        # Detectar número de caras en la imagen
        if results.detections:
            # Aislar cada cara y realizar el proceso HRE
            for idx, face in enumerate(results.detections):
                c_image = image.copy()

                height, width, channels = c_image.shape # Webcam 480 x 640 x 3
                bounding_box = face.location_data.relative_bounding_box

                # Rectángulo cara
                x = int(bounding_box.xmin * width) - margin
                w = x + int(bounding_box.width * width) + 2 * margin
                y = int(bounding_box.ymin * height) - 2 * margin
                h = y + int(int((bounding_box.width * width) + 2 * margin) * 1.25)

                if x < 0: x = 0
                if w > width: w = width
                if y < 0: y = 0
                if h > height: break

                cv2.rectangle(image, (x, y), (w, h), (0, 255, 0), 1)

                # Se sombrean las zonas donde no esté la cara y el cuerpo que se analizan
                width_head = w - x

                left_lim = int(x - 0.55 * width_head)
                if left_lim < 0: left_lim = 0
                cv2.rectangle(c_image, (0, 0), (left_lim, height), (0, 0, 0), -1)

                right_lim = int(w + 0.55 * width_head)
                if right_lim > width: right_lim = width
                cv2.rectangle(c_image, (right_lim, 0), (width, height), (0, 0, 0), -1)

                # Recorte de la cara para reconocimiento facial
                image_crop = image[y:h, x:w]

                # Localizamos posición para realizar seguimiento
                head_rotX, head_rotY, head_rotZ, [nose_start, nose_end], open_mouth = dp.head_pose_zone(c_image, face_mesh)

                # Dibujar proyección cara en la imagen
                if not head_rotX == None and not head_rotY == None and not head_rotZ == None:
                    if printed_nose:
                        # Mostrar dirección nariz
                        cv2.line(image, nose_start, nose_end, (255, 0, 0), 2)
                        # Mostramos comentario orientación en la imagen
                        sr.head_text_view(head_rotX, head_rotY, head_rotZ, image, [x, y, w, h])
                        sr.head_text_pose(head_rotX, head_rotY, head_rotZ, image, [x, y, w, h])

                # Reconocimiento facial
                id_face = "desconocido"

                # Comparamos con una posición existente anterior de caras conocidas (Seguimiento)
                if not nose_start == None:
                    for id, last_pos_nose in last_nose.items():
                        if dp.aprox_nose(nose_start, last_pos_nose):
                            id_face = id

                # Si no es una cara localizada y se tiene la nariz se busca en el seguimiento de los desconocidos
                # Este seguimiento se realiza en periodos de 1 segundo y se vuelve a probar el reconocmiento
                if id_face == "desconocido" and not nose_start == None:
                    desc_del = []
                    locate_unknown = False
                    for index, [pos_nose, last_time] in enumerate(desc):
                        if start_time - last_time > 1:
                            desc_del.append(index)
                        elif dp.aprox_nose(nose_start, pos_nose):
                            desc[index] = [nose_start, last_time]
                            locate_unknown = True
                    for line_del in reversed(desc_del):
                        desc.pop(line_del)

                    if not locate_unknown:
                        id_face = f2.id_detect(image_crop, known_face_encodings, known_face_ids)
                        if id_face == "desconocido":
                            desc.append([nose_start, start_time])
                    else:
                        locate_unknown = False

                # Si la cara es conocida y se tiene la nariz para realizar seguimiento
                if not id_face == "desconocido" and not nose_start == None:
                    last_nose[id_face] = nose_start
                    if printed_id:
                        sr.show_id_text(image, [x, y, w, h], id_face)

                    if not head_rotX == None and not head_rotY == None and not head_rotZ == None:
                        # Reconocimiento de emociones
                        emotion_face = DeepFace.analyze(img_path = image_crop,
                                                   actions = ['emotion'],
                                                   models = None,
                                                   enforce_detection = False,
                                                   detector_backend = backends[5],
                                                   prog_bar = False)

                        # Si cara está en la parte central de la imagen se asume que las pulsaciones del teclado pueden ser suyas
                        pulse_center = None
                        total_pulse = 0
                        if pulse and width / 3 < x + (w - x) / 2 < 2 * width / 3:
                            pulse_center = True
                            total_pulse += 1

                        # Media de emociones y las pulsaciones del teclado en los últimos 3 segundos aproximadamente
                        dict_emotion = {"enfado": 0, "disgusto": 0, "miedo": 0, "feliz": 0, "Triste": 0, "Sorpresa": 0,
                                        "neutral": 0}
                        rev_array_memory = list(reversed(array_memory))
                        array_memory_filter = list(filter(lambda col: col[1] == id_face, rev_array_memory))
                        array_memory_filter = array_memory_filter[0:18:1]

                        dict_emotion[emotion_dict[emotion_face['dominant_emotion']]] += 1
                        for st, id, em, hrY, hrX, brY, om, pc, emf, ppc, dnmf, dnmf2 in array_memory_filter:
                            dict_emotion[em] += 1
                            if pc == True:
                                total_pulse += 1

                        emotion_max_frec = max(dict_emotion, key=dict_emotion.get)
                        por_pulse_center = total_pulse / (len(array_memory_filter)+1)

                        # Mostrar emoción media
                        if emotion_face:
                            sr.show_emotion_text(emotion_max_frec, (x, y, w, h), image)

                        # Obtener datos del cuerpo
                        body_rotY, [p1, p2] = dp.body_pose_zone(c_image, pose, nose_end[0]-nose_start[0])

                        # Pintar datos del cuerpo en imagen
                        if printed_body and not body_rotY == None:
                            # Mostrar dirección torso
                            cv2.circle(image, p1, radius=5, color=(0, 0, 255), thickness=5)
                            cv2.line(image, p1, p2, (255, 0, 0), 2)
                            # Mostramos comentario orientación en la imagen
                            sr.body_text_pose(body_rotY, image, [x, y, w, h])

                    # Predicción HRE con SVM
                    testX = np.array([np.float32(head_rotY), np.float32(head_rotX), np.float32(body_rotY), np.float32(open_mouth), np.float32(por_pulse_center)])
                    resp = svm.predict(np.array([testX], dtype=np.float32))[1]

                    # Registramos a las personas nuevas
                    if not id_face in atencion_dict:
                        atencion_dict[id_face] = []
                    atencion_dict[id_face].append(name_array_video[int(resp[0][0])])

                    # Hacemos la media de las últimas 10 imágenes para dar resultados estables
                    if len(atencion_dict[id_face]) > 10:
                        atencion_dict[id_face].pop(0)

                    dict_name = {"muy atento": 0, "atento": 0, "poco atento": 0, "despistado": 0, "muy despistado": 0}
                    for name in atencion_dict[id_face]:
                        dict_name[name] += 1
                    dict_name_max_frec = max(dict_name, key=dict_name.get)

                    # Predicción HRE con Keras NN
                    testX2 = [[np.float32(head_rotY), np.float32(head_rotX), np.float32(body_rotY), np.float32(open_mouth), np.float32(por_pulse_center)]]
                    resp2 = keras_model.predict(np.array(testX2), verbose=0)
                    resp2 = np.argmax(resp2[0])

                    # Registramos a las personas nuevas en el seguimiento de la IA de keras
                    if not id_face in atencion_dict_keras:
                        atencion_dict_keras[id_face] = []
                    atencion_dict_keras[id_face].append(name_array_video[int(np.round(resp[-1]))])

                    # Hacemos la media de las últimas 10 imágenes para dar resultados estables
                    if len(atencion_dict_keras[id_face]) > 10:
                        atencion_dict_keras[id_face].pop(0)

                    dict_name2 = {"muy atento": 0, "atento": 0, "poco atento": 0, "despistado": 0, "muy despistado": 0}
                    for name in atencion_dict_keras[id_face]:
                        dict_name2[name] += 1
                    dict_name_max_frec2 = max(dict_name2, key=dict_name2.get)

                    # Mostrar atención media
                    if printed_engagement:
                        sr.show_engagement_text(dict_name_max_frec2, (x, y, w, h), image)

                    # Almacenamos todos los datos
                    array_memory.append(
                        [start_time, id_face, emotion_dict[emotion_face['dominant_emotion']], head_rotY, head_rotX,
                         body_rotY, open_mouth, pulse_center, emotion_max_frec, por_pulse_center, dict_name_max_frec, dict_name_max_frec2])

                    if printed_array:
                        print(array_memory[-1])
                        # print(dict_name_max_frec)
                pulse = None

        #############################
        end_time = time.time()
        total_time = end_time - start_time

        fps = 1/total_time
        print("FPS: ", fps)

        sr.print_FPS(fps, image)

        cv2.imshow('Posicion cara y cuerpo', image)

        if cv2.waitKey(5) & 0xFF == 27:
            break

    # Liberamos el recurso de la cámara y cerramos todas las ventanas
    cap.release()
    cv2.destroyAllWindows()

    # Creamos tantos gráficos como personas se han reconocido
    id_faces = last_nose.keys()
    fig, axis = plt.subplots(len(id_faces), 2)

    # Mostramos gráficos en función de las personas que han aparecido en la interacción y se comparan las dos IA entrenadas.
    for index, id in enumerate(id_faces):
        array_memory_filter = list(filter(lambda col: col[1] == id, array_memory))
        st_time = []
        emot = []
        eng = []
        eng2 = []

        first_time = array_memory_filter[0][0]

        for idx, [st, id, em, hrY, hrX, brY, om, pc, emf, ppc, dnmf, dnmf2] in enumerate(array_memory_filter):
            st_time.append(int((st - first_time) * 1000))
            emot.append(em)
            eng.append(4 - name_array_video.index(dnmf))
            eng2.append(4 - name_array_video.index(dnmf2))

            if (idx+1) < len(array_memory_filter):
                if array_memory_filter[idx+1][0] - st > 1:
                    st_time.append(int((st + 0.45 - first_time) * 1000))
                    emot.append(None)
                    eng.append(0)
                    eng2.append(0)

                    st_time.append(int((array_memory_filter[idx+1][0] - 0.45 - first_time) * 1000))
                    emot.append(None)
                    eng.append(0)
                    eng2.append(0)

        if len(id_faces) > 1:
            axis[index, 0].set_title("Interacción " + id + " (SVM)")
            axis[index, 0].axis([None, None, 0, 5])
            axis[index, 0].plot(st_time, eng, 'r')

            axis[index, 1].set_title("Interacción " + id + " (NN)")
            axis[index, 1].axis([None, None, 0, 5])
            axis[index, 1].plot(st_time, eng2, 'g')
        else:
            axis[0].set_title("Interacción " + id + " (SVM)")
            axis[0].axis([None, None, 0, 5])
            axis[0].plot(st_time, eng, 'r')

            axis[1].set_title("Interacción " + id + " (NN)")
            axis[1].axis([None, None, 0, 5])
            axis[1].plot(st_time, eng2, 'g')

    plt.suptitle("Interacción Humano-Robot")
    plt.xlabel("Milisegundos")
    plt.subplots_adjust(hspace=0.5)
    plt.show()
