# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import mediapipe as mp
import face_recognition
from deepface import DeepFace
import os
import time
from pynput import keyboard

import pandas as pd
from pandas import ExcelWriter
import numpy as np

import detect_pose as dp
import face as f2
import show_results as sr

global pulse

# Función para registrar pulsaciones
def on_press(key):
    global pulse
    pulse = True

if __name__ == '__main__':
    name_array_video = ["muy atento", "atento", "poco atento", "despistado", "muy despistado"]
    pos = 0

    # Ubicación donde se guardarán los videos y los excel
    # name_video = "recursos/" + name_array_video[pos].replace(" ", "_") + ".wmv"
    # name_excel = "recursos/excel_" + name_array_video[pos].replace(" ", "_") + ".xlsx"
    name_video = "recursos_testeo/" + name_array_video[pos].replace(" ", "_") + ".wmv"
    name_excel = "recursos_testeo/excel_" + name_array_video[pos].replace(" ", "_") + ".xlsx"

    # Listener para detectar pulsaciones
    pulse = None
    listener = keyboard.Listener(on_press=on_press)
    listener.start()

    # ¿Pintar imagen con información?
    printed_id = True
    printed_nose = True
    printed_emotion = True
    printed_body = True
    printed_array = False

    # Objetos detectar cara
    mp_face_detection = mp.solutions.face_detection
    face_detector = mp_face_detection.FaceDetection(min_detection_confidence=0.6)

    # Objetos identificar cara
    known_face_encodings = []
    known_face_ids = []
    last_nose = {}

    for id_image in os.listdir("../facial_recognition/known_images"):
        image = face_recognition.load_image_file("../facial_recognition/known_images/" + id_image)
        known_face_encodings.append(face_recognition.face_encodings(image)[0])
        known_face_ids.append(os.path.splitext(id_image)[0])

    # Objetos detectar emociones
    backends = ['opencv', 'ssd', 'dlib', 'mtcnn', 'retinaface', 'mediapipe']
    emotion_dict = {"angry": "enfado",
                    "disgust": "disgusto",
                    "fear": "miedo",
                    "happy": "feliz",
                    "sad": "Triste",
                    "surprise": "Sorpresa",
                    "neutral": "neutral"}

    # Objetos cara
    mp_face_mesh = mp.solutions.face_mesh
    face_mesh = mp_face_mesh.FaceMesh(max_num_faces=3, min_detection_confidence=0.5, min_tracking_confidence=0.5)

    # Objetos torso
    mp_pose = mp.solutions.pose
    pose = mp_pose.Pose(min_detection_confidence=0.5, min_tracking_confidence=0.5)

    # Array para almacenar información [instante, id_face, emotion_face, head_rotY, head_rotX, body_rotY, pulse]
    array_memory = []
    desc = []
    margin = 20

    # Inicialización cámara
    cap = cv2.VideoCapture(0)
    success, image = cap.read()
    height, width, channels = image.shape

    # Características del video
    video = cv2.VideoWriter(name_video, cv2.VideoWriter_fourcc(*'mp4v'), 2, (width, height))

    while cap.isOpened():
        start_time = time.time()
        #############################
        # Capturar imagen y procesamiento
        success, image = cap.read()

        image = cv2.flip(image, 1)

        results = face_detector.process(image)

        if results.detections:
            for idx, face in enumerate(results.detections):
                c_image = image.copy()

                height, width, channels = c_image.shape # Webcam 480 x 640 x 3
                bounding_box = face.location_data.relative_bounding_box

                # Rectángulo cara
                x = int(bounding_box.xmin * width) - margin
                w = x + int(bounding_box.width * width) + 2 * margin
                y = int(bounding_box.ymin * height) - 2 * margin
                h = y + int(int((bounding_box.width * width) + 2 * margin) * 1.25)

                if x < 0: x = 0
                if w > width: w = width
                if y < 0: y = 0
                if h > height: break

                cv2.rectangle(image, (x, y), (w, h), (0, 255, 0), 1)

                # límites de sombreado
                width_head = w - x

                left_lim = int(x - 0.55 * width_head)
                if left_lim < 0: left_lim = 0
                cv2.rectangle(c_image, (0, 0), (left_lim, height), (0, 0, 0), -1)

                right_lim = int(w + 0.55 * width_head)
                if right_lim > width: right_lim = width
                cv2.rectangle(c_image, (right_lim, 0), (width, height), (0, 0, 0), -1)

                image_crop = image[y:h, x:w]
                head_rotX, head_rotY, head_rotZ, [nose_start, nose_end], open_mouth = dp.head_pose_zone(c_image, face_mesh)


                if not head_rotX == None and not head_rotY == None and not head_rotZ == None:
                    if printed_nose:
                        # Mostrar dirección nariz
                        cv2.line(image, nose_start, nose_end, (255, 0, 0), 2)
                        # Mostramos comentario orientación en la imagen
                        sr.head_text_view(head_rotX, head_rotY, head_rotZ, image, (x, y, w, h))
                        sr.head_text_pose(head_rotX, head_rotY, head_rotZ, image, (x, y, w, h))

                id_face = "desconocido"

                # Buscar localización de caras conocidas
                if not nose_start == None:
                    for id, last_pos_nose in last_nose.items():
                        if dp.aprox_nose(nose_start, last_pos_nose):
                            id_face = id

                # Buscar y guardar localización de caras desconocidas
                if id_face == "desconocido" and not nose_start == None:
                    desc_del = []
                    locate_unknown = False
                    for index, [pos_nose, last_time] in enumerate(desc):
                        if start_time - last_time > 1:
                            desc_del.append(index)
                        elif dp.aprox_nose(nose_start, pos_nose):
                            desc[index] = [nose_start, last_time]
                            locate_unknown = True

                    for line_del in reversed(desc_del):
                        desc.pop(line_del)

                    if not locate_unknown:
                        id_face = f2.id_detect(image_crop, known_face_encodings, known_face_ids)
                        if id_face == "desconocido":
                            desc.append([nose_start, start_time])
                    else:
                        locate_unknown = False

                # Análisis de caras conocidas
                if not id_face == "desconocido" and not nose_start == None:
                    last_nose[id_face] = nose_start
                    if printed_id:
                        sr.show_id_text(image, [x, y, w, h], id_face)

                    if not head_rotX == None and not head_rotY == None and not head_rotZ == None:
                        if printed_nose:
                            # Mostrar dirección nariz
                            cv2.line(image, nose_start, nose_end, (255, 0, 0), 2)
                            # Mostramos comentario orientación en la imagen
                            sr.head_text_view(head_rotX, head_rotY, head_rotZ, image, (x, y, w, h))
                            sr.head_text_pose(head_rotX, head_rotY, head_rotZ, image, (x, y, w, h))

                        emotion_face = DeepFace.analyze(img_path = image_crop,
                                                   actions = ['emotion'],
                                                   models = None,
                                                   enforce_detection = False,
                                                   detector_backend = backends[5],
                                                   prog_bar = False)

                        # Si cara está en la parte central de la imagen se asume que las pulsaciones pueden ser suyas
                        pulse_center = None
                        total_pulse = 0
                        if pulse and width / 3 < x + (w - x) / 2 < 2 * width / 3:
                            pulse_center = True
                            total_pulse += 1

                        dict_emotion = {"enfado": 0, "disgusto": 0, "miedo": 0, "feliz": 0, "Triste": 0, "Sorpresa": 0,
                                        "neutral": 0}

                        rev_array_memory = list(reversed(array_memory))
                        array_memory_filter = list(filter(lambda col: col[1] == id_face, rev_array_memory))
                        array_memory_filter = array_memory_filter[0:18:1]

                        dict_emotion[emotion_dict[emotion_face['dominant_emotion']]] += 1
                        for st, id, em, hrY, hrX, brY, om, pc, emf, ppc in array_memory_filter:
                            dict_emotion[em] += 1
                            if pc == True:
                                total_pulse += 1

                        emotion_max_frec = max(dict_emotion, key=dict_emotion.get)
                        por_pulse_center = total_pulse / (len(array_memory_filter)+1)

                        if emotion_face:
                            sr.show_emotion_text(emotion_max_frec, (x, y, w, h), image)

                        body_rotY, [p1, p2] = dp.body_pose_zone(c_image, pose, nose_end[0]-nose_start[0])

                        if printed_body and not body_rotY == None:
                            # Mostrar dirección torso
                            cv2.circle(image, p1, radius=5, color=(0, 0, 255), thickness=5)
                            cv2.line(image, p1, p2, (255, 0, 0), 2)
                            # Mostramos comentario orientación en la imagen
                            sr.body_text_pose(body_rotY, image, [x, y, w, h])

                    array_memory.append([start_time, id_face, emotion_dict[emotion_face['dominant_emotion']], head_rotY, head_rotX, body_rotY, open_mouth, pulse_center, emotion_max_frec, por_pulse_center])
                    if printed_array:
                        print(array_memory[-1])

                    video.write(image)

                # Actualiza a None las pulsaciones en cada ciclo
                pulse = None

        cv2.imshow('Posicion cara y cuerpo', image)

        if cv2.waitKey(5) & 0xFF == 27:
            break

    # Dar formato a los datos para guardar en el excel
    array_memory_np = np.array(array_memory)

    dataFrame = pd.DataFrame({'start_time': array_memory_np[20:, 0],
                       'id_face': array_memory_np[20:, 1],
                       'emotion': array_memory_np[20:, 2],
                       'head_rotY': array_memory_np[20:, 3],
                       'head_rotX': array_memory_np[20:, 4],
                       'body_rotY': array_memory_np[20:, 5],
                       'open_mouth': array_memory_np[20:, 6],
                       'pulse_center': array_memory_np[20:, 7],
                       'emotion_max_frec': array_memory_np[20:, 8],
                       'por_pulse_center': array_memory_np[20:, 9]})

    dataFrame = dataFrame[['start_time', 'id_face', 'emotion', 'head_rotY', 'head_rotX', 'body_rotY', 'open_mouth', 'pulse_center', 'emotion_max_frec', 'por_pulse_center']]
    writer = ExcelWriter(name_excel)
    dataFrame.to_excel(writer, name_array_video[pos], index=False)
    writer.save()

    video.release()
    cap.release()
    cv2.destroyAllWindows()