# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import pandas as pd
import numpy as np
import os
import math
import time

# Función que almacena todas las combinaciones de las secciones posible
def iteration_pos_num():
    global num
    global prof
    global n_bloques
    global comb

    if prof < n_bloques - 1:
        for i in range(n_bloques):
            num[prof] = i
            prof += 1
            iteration_pos_num()
            prof -= 1

    else:
        for i in range(n_bloques):
            num[prof] = i
            if len(set(num)) == n_bloques:
                comb.append(list(num))
                # text = "".join([str(_) for _ in num])
                # print(text)

# Transforma segundos a horas, minutos y segundos
def s2hms(seg):
    h = int(seg / 60 / 60)
    seg -= h*60*60
    m = int(seg/60)
    seg -= m*60
    print("Tiempo previsto para finalizar: {horas} horas, {minutos} minutos y {segundos} segundos".format(horas=h, minutos=m, segundos=int(seg)))

prof = 0

# Número de divisiones
n_bloques = 1 # Define numero de bloques "comb = n^n - (repetidos)"

num = np.zeros(n_bloques, dtype=int)
comb = []

print("Calculando combinaciones")
iteration_pos_num()
print("{n_comb} combinaciones calculadas".format(n_comb=str(len(comb))))

# Inicializamos objetos y cargamos datos en objetos
name_array_video = ["muy atento", "atento", "poco atento", "despistado", "muy despistado"]
exist_all = True

for name in name_array_video:
    if not os.path.exists("recursos/excel_" + name.replace(" ", "_") + ".xlsx"):
        exist_all = False

descriptores = [] # trainX
trainY = []

if exist_all:
    for index, name in enumerate(name_array_video):
        dataFrame = pd.read_excel("recursos/excel_" + name.replace(" ", "_") + ".xlsx")
        head_rotY_excel = dataFrame['head_rotY'].tolist()
        head_rotX_excel = dataFrame['head_rotX'].tolist()
        body_rotY = dataFrame['body_rotY'].tolist()
        open_mouth_excel = dataFrame['open_mouth'].tolist()
        por_pulse_center_excel = dataFrame['por_pulse_center'].tolist()

        for i in range(len(dataFrame)):
            descriptores.append([np.float32(head_rotY_excel[i]), np.float32(head_rotX_excel[i]), np.float32(body_rotY[i]), np.float32(open_mouth_excel[i]), np.float32(por_pulse_center_excel[i])])
        trainY.extend([index] * len(dataFrame))

        # Datos validación
        exist_all_test = True

        for name in name_array_video:
            if not os.path.exists("recursos_testeo/excel_" + name.replace(" ", "_") + ".xlsx"):
                exist_all_test = False

        testX = []
        testY = []

        if exist_all_test:
            for index, name in enumerate(name_array_video):
                dataFrame = pd.read_excel("recursos_testeo/excel_" + name.replace(" ", "_") + ".xlsx")
                head_rotY_excel = dataFrame['head_rotY'].tolist()
                head_rotX_excel = dataFrame['head_rotX'].tolist()
                body_rotY = dataFrame['body_rotY'].tolist()
                open_mouth_excel = dataFrame['open_mouth'].tolist()
                por_pulse_center_excel = dataFrame['por_pulse_center'].tolist()

                for i in range(len(dataFrame)):
                    testX.append(
                        [np.float32(head_rotY_excel[i]), np.float32(head_rotX_excel[i]), np.float32(body_rotY[i]),
                         np.float32(open_mouth_excel[i]), np.float32(por_pulse_center_excel[i])])
                testY.extend([index] * len(dataFrame))

        testX = np.array(testX)

    # Entrenar y validar cada combinación
    div = math.ceil(len(descriptores) / n_bloques)
    bloques_descriptores = [descriptores[i:i + div] for i in range(0, len(descriptores), div)]
    bloques_trainY = [trainY[i:i + div] for i in range(0, len(trainY), div)]

    print("Entrenando y validando modelos...")

    combinacion_mas_precisa = {"n_bloques": n_bloques, "combinacion": 0, "precision": 0, "validez": 0}
    combinacion_mas_valida = {"n_bloques": n_bloques, "combinacion": 0, "precision": 0, "validez": 0}

    svm = cv2.ml.SVM_create()
    svm.setType(cv2.ml.SVM_C_SVC)
    svm.setKernel(cv2.ml.SVM_RBF)

    for index, position in enumerate(comb):
        start_time = time.time()
        descriptores_comb = []
        trainY_comb = []

        for pos_num in position:
            descriptores_comb.extend(bloques_descriptores[pos_num])
            trainY_comb.extend(bloques_trainY[pos_num])

        descriptores_comb = np.array(descriptores_comb)
        trainY_comb = np.array(trainY_comb)

        # Entrenamos SVM
        svm.trainAuto(descriptores_comb, cv2.ml.ROW_SAMPLE, trainY_comb)
        svm.save("recursos/result_model.xml")
        # svm = cv2.ml.SVM_load("recursos/result_model_precision.xml")
        aciertos_validos = []
        aciertos_precisos = []

        for ind, linea in enumerate(testX):
            resp = svm.predict(np.array([linea], dtype=np.float32))[1]
            if int(resp[0][0]) - 1 <= testY[ind] <= int(resp[0][0]) + 1:
                aciertos_validos.append(True)
            else:
                aciertos_validos.append(False)

            if testY[ind] == int(resp[0][0]):
                aciertos_precisos.append(True)
            else:
                aciertos_precisos.append(False)

        porc_validos = 100 * aciertos_validos.count(True) / len(aciertos_validos)
        porc_preciso = 100 * aciertos_precisos.count(True) / len(aciertos_precisos)

        print('Combinación {position} testeada'.format(position=str(position)))

        actualizar = False

        if combinacion_mas_precisa["precision"] < porc_preciso:
            combinacion_mas_precisa["combinacion"] = position
            combinacion_mas_precisa["precision"] = porc_preciso
            combinacion_mas_precisa["validez"] = porc_validos
            svm.save("recursos/result_model_precision.xml")
            actualizar = True

        if combinacion_mas_valida["validez"] < porc_validos:
            combinacion_mas_valida["combinacion"] = position
            combinacion_mas_valida["precision"] = porc_preciso
            combinacion_mas_valida["validez"] = porc_validos
            svm.save("recursos/result_model_validez.xml")
            actualizar = True

        if actualizar:
            mejor_resultado = open("mejor_resultado.txt", "w+")
            str1 = "Combinacion mas precisa: {position}\n\tvalidez: {porc_validos}\n\tprecision: {porc_preciso}\n\n".format(position=combinacion_mas_precisa["combinacion"], porc_preciso=round(combinacion_mas_precisa["precision"], 2), porc_validos=round(combinacion_mas_precisa["validez"], 2))
            str2 = "Combinacion mas valida: {position}\n\tvalidez: {porc_validos}\n\tprecision: {porc_preciso}".format(position=combinacion_mas_valida["combinacion"], porc_preciso=round(combinacion_mas_valida["precision"], 2), porc_validos=round(combinacion_mas_valida["validez"], 2))
            mejor_resultado.writelines([str1, str2])
            mejor_resultado.close()

        num = index/100
        if num.is_integer():
            ultimo_resultado = open("ultimo_resultado_analizado.txt", "w+")
            ultimo_resultado.write("Ultimo resultado analizado: {index}\ncombinacion: {position}".format(position=combinacion_mas_precisa["combinacion"], index=index))
            ultimo_resultado.close()

        end_time = time.time()
        calc_time = end_time - start_time

        s2hms((len(comb) - (index + 1)) * calc_time)

    print('Combinación más precisa:\n\tCombinacion: {position}\n\tResultados validos: {porc_validos} %\n\tResultados precisos: {porc_preciso} %'.format(position = str(combinacion_mas_precisa["combinacion"]), porc_validos=round(combinacion_mas_precisa["validez"], 2), porc_preciso=round(combinacion_mas_precisa["precision"], 2)))
    print('Combinación más valida:\n\tCombinacion: {position}\n\tResultados validos: {porc_validos} %\n\tResultados precisos: {porc_preciso} %'.format(position = str(combinacion_mas_valida["combinacion"]), porc_validos=round(combinacion_mas_valida["validez"], 2), porc_preciso=round(combinacion_mas_valida["precision"], 2)))
