# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import pandas as pd
import numpy as np
import os
import math
import time

import keras
import sklearn
from keras.models import Sequential
from keras.layers.core import Dense
from keras.layers.core import Dropout

import tensorflow as tf
import tensorboard
import datetime

import matplotlib.pyplot as plt
import seaborn as sns

sns.set_theme()

# Ayuda a función para ordenar lista según el tercer parámetro
def order_list(e):
  return e[2]

# Función que almacena todas las combinaciones de las secciones posible
def iteration_pos_num():
    global num
    global prof
    global n_bloques
    global comb

    if prof < n_bloques - 1:
        for i in range(n_bloques):
            num[prof] = i
            prof += 1
            iteration_pos_num()
            prof -= 1

    else:
        for i in range(n_bloques):
            num[prof] = i
            if len(set(num)) == n_bloques:
                comb.append(list(num))
                # text = "".join([str(_) for _ in num])
                # print(text)

# Transforma segundos a horas, minutos y segundos
def s2hms(seg):
    h = int(seg / 60 / 60)
    seg -= h*60*60
    m = int(seg/60)
    seg -= m*60
    print("Tiempo previsto para finalizar: {horas} horas, {minutos} minutos y {segundos} segundos".format(horas=h, minutos=m, segundos=int(seg)))

# Función para entrenar modelos
def entrenamiento_parametros(model, epochs):
    div = math.ceil(len(descriptores) / n_bloques)
    bloques_descriptores = [descriptores[i:i + div] for i in range(0, len(descriptores), div)]
    bloques_trainY = [trainY[i:i + div] for i in range(0, len(trainY), div)]

    print("Entrenando y validando modelos...")

    for index, position in enumerate(comb):
        descriptores_comb = []
        trainY_comb = []

        for pos_num in position:
            descriptores_comb.extend(bloques_descriptores[pos_num])
            trainY_comb.extend(bloques_trainY[pos_num])

        descriptores_comb = np.array(descriptores_comb, dtype=np.float32)
        trainY_comb = np.array(trainY_comb, dtype=np.float32)

        # Entrenamos red neuronal
        checkpoint_filepath = 'recursos/result_model_precision3_keras.h5'
        model_checkpoint_callback = keras.callbacks.ModelCheckpoint(
            filepath=checkpoint_filepath,
            monitor='val_accuracy',
            mode='max',
            save_best_only=True)

        descriptores_comb, trainY_comb = sklearn.utils.shuffle(descriptores_comb, trainY_comb)

        log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)

        model.fit(descriptores_comb, trainY_comb, batch_size=8, verbose=2, epochs=epochs, callbacks=[model_checkpoint_callback, tensorboard_callback], validation_data=(testX, testY), validation_split=0.2)

        model = keras.models.load_model('recursos/result_model_precision3_keras.h5', compile=True)
        results = model.evaluate(testX, testY)
    return results


prof = 0

# Número de divisiones
n_bloques = 1  # Define numero de bloques "comb = n^n - (repetidos)"

num = np.zeros(n_bloques, dtype=int)
comb = []

# Inicializamos objetos y cargamos datos
name_array_video = ["muy atento", "atento", "poco atento", "despistado", "muy despistado"]
exist_all = True

for name in name_array_video:
    if not os.path.exists("recursos/excel_" + name.replace(" ", "_") + ".xlsx"):
        exist_all = False

descriptores = []  # trainX
trainY = []

if exist_all:
    for index, name in enumerate(name_array_video):
        dataFrame = pd.read_excel("recursos/excel_" + name.replace(" ", "_") + ".xlsx")
        head_rotY_excel = dataFrame['head_rotY'].tolist()
        head_rotX_excel = dataFrame['head_rotX'].tolist()
        body_rotY = dataFrame['body_rotY'].tolist()
        open_mouth_excel = dataFrame['open_mouth'].tolist()
        por_pulse_center_excel = dataFrame['por_pulse_center'].tolist()

        for i in range(len(dataFrame)):
            descriptores.append(
                [np.float32(head_rotY_excel[i]), np.float32(head_rotX_excel[i]), np.float32(body_rotY[i]),
                 np.float32(open_mouth_excel[i]), np.float32(por_pulse_center_excel[i])])
        trainY.extend([index] * len(dataFrame))

        # Datos validación
        exist_all_test = True

        for name in name_array_video:
            if not os.path.exists("recursos_testeo/excel_" + name.replace(" ", "_") + ".xlsx"):
                exist_all_test = False

        testX = []
        testY = []

        if exist_all_test:
            for index, name in enumerate(name_array_video):
                dataFrame = pd.read_excel("recursos_testeo/excel_" + name.replace(" ", "_") + ".xlsx")
                head_rotY_excel = dataFrame['head_rotY'].tolist()
                head_rotX_excel = dataFrame['head_rotX'].tolist()
                body_rotY = dataFrame['body_rotY'].tolist()
                open_mouth_excel = dataFrame['open_mouth'].tolist()
                por_pulse_center_excel = dataFrame['por_pulse_center'].tolist()

                for i in range(len(dataFrame)):
                    testX.append(
                        [np.float32(head_rotY_excel[i]), np.float32(head_rotX_excel[i]), np.float32(body_rotY[i]),
                         np.float32(open_mouth_excel[i]), np.float32(por_pulse_center_excel[i])])
                testY.extend([index] * len(dataFrame))

        testX = np.array(testX, dtype=np.float32)
        testY = np.array(testY, dtype=np.float32)

# Calculamos combinaciones
print("Calculando combinaciones")
iteration_pos_num()
print("{n_comb} combinaciones calculadas".format(n_comb=str(len(comb))))

activations = ['relu', 'sigmoid', 'softmax', 'softplus', 'softsign', 'tanh', 'selu', 'elu', 'exponential']
optimicers = ['SGD', 'RMSprop', 'Adam', 'Adadelta', 'Adagrad', 'Adamax', 'Nadam', 'Ftrl']
probabilistic_losses = ['kl_divergence', 'binary_crossentropy', 'poisson', 'categorical_crossentropy', 'binary_crossentropy']
regression_losses = ['mean_squared_error', 'mean_absolute_error', 'mean_absolute_percentage_error', 'mean_squared_logarithmic_error', 'cosine_similarity', 'huber_loss', 'log_cosh']
metrics = [['accuracy'], ['categorical_accuracy'], ['sparse_categorical_accuracy'], ['top_k_categorical_accuracy'], ['mean_squared_error'], ['mean_absolute_error'], ['mean_absolute_percentage_error']]

combinaciones_parametros = []

# # BUSQUEDA MEJORES PARAMETROS ACTIVATIONS
# for idx1, param_in in enumerate(activations):
#     for idx2, param_out in enumerate(activations):
#         start_time = time.time()
#
#         model = Sequential()
#         model.add(Dense(16, input_dim=5, activation=param_in))
#         model.add(Dense(1, activation=param_out))
#
#         model.compile(loss='huber_loss',
#                       optimizer='adam',
#                       metrics=['categorical_accuracy'])
#         epochs = 500
#         best_result = entrenamiento_parametros(model, epochs)
#
#         end_time = time.time()
#         calc_time = end_time - start_time
#
#         s2hms((len(activations)**2 - (idx1 + 1)*(idx2 + 1)) * calc_time)
#
#         combinaciones_parametros.append([param_in, param_out, best_result["combinacion"], round(best_result["precision"], 2)])
#         print('Combinación más precisa:\n\tparam in: {param_in}\n\tparam out: {param_out}\n\tCombinacion: {position}\n\tResultados precisos: {porc_preciso} %'.format(
#                 param_in=param_in,
#                 param_out=param_out,
#                 position=str(best_result["combinacion"]),
#                 porc_preciso=round(best_result["precision"], 2)))

#
# # BUSQUEDA MEJORES PARAMETROS OPTIMICERS
# for idx, param_opt in enumerate(optimicers):
#     start_time = time.time()
#
#     model = Sequential()
#     model.add(Dense(16, input_dim=5, activation='selu'))
#     model.add(Dense(1, activation='elu'))
#
#     model.compile(loss='huber_loss',
#                   optimizer=param_opt,
#                   metrics=['categorical_accuracy'])
#     epochs = 500
#     best_result = entrenamiento_parametros(model, epochs)
#
#     end_time = time.time()
#     calc_time = end_time - start_time
#
#     s2hms((len(param_opt) - (idx + 1)) * calc_time)
#
#     combinaciones_parametros.append([param_opt, best_result["combinacion"], round(best_result["precision"], 2)])
#     print('Combinación más precisa:\n\tParametro optimicer: {param_opt}\n\tCombinacion: {position}\n\tResultados precisos: {porc_preciso} %'.format(
#             param_opt=param_opt,
#             position=str(best_result["combinacion"]),
#             porc_preciso=round(best_result["precision"], 2)))

#
# # BUSQUEDA MEJORES PARAMETROS PROBABILISTIC LOSSES
# for idx, param_prob_loss in enumerate(probabilistic_losses):
#     start_time = time.time()
#
#     model = Sequential()
#     model.add(Dense(16, input_dim=5, activation='selu'))
#     model.add(Dense(1, activation='elu'))
#
#     model.compile(loss=param_prob_loss,
#                   optimizer='adam',
#                   metrics=['categorical_accuracy'])
#     epochs = 500
#     best_result = entrenamiento_parametros(model, epochs)
#
#     end_time = time.time()
#     calc_time = end_time - start_time
#
#     s2hms((len(param_prob_loss) - (idx + 1)) * calc_time)
#
#     combinaciones_parametros.append([param_prob_loss, best_result["combinacion"], round(best_result["precision"], 2)])
#     print('Combinación más precisa:\n\tParametro Probabilistic Losses: {param_prob_loss}\n\tCombinacion: {position}\n\tResultados precisos: {porc_preciso} %'.format(
#             param_prob_loss=param_prob_loss,
#             position=str(best_result["combinacion"]),
#             porc_preciso=round(best_result["precision"], 2)))


# # BUSQUEDA MEJORES PARAMETROS REGRESSION LOSSES
# for idx, param_reg_loss in enumerate(regression_losses):
#     start_time = time.time()
#
#     model = Sequential()
#     model.add(Dense(16, input_dim=5, activation='selu'))
#     model.add(Dense(1, activation='elu'))
#
#     model.compile(loss=param_reg_loss,
#                   optimizer='adam',
#                   metrics=['categorical_accuracy'])
#     epochs = 500
#     best_result = entrenamiento_parametros(model, epochs)
#
#     end_time = time.time()
#     calc_time = end_time - start_time
#
#     s2hms((len(param_reg_loss) - (idx + 1)) * calc_time)
#
#     combinaciones_parametros.append([param_reg_loss, best_result["combinacion"], round(best_result["precision"], 2)])
#     print('Combinación más precisa:\n\tParametro Regression Losses: {param_reg_loss}\n\tCombinacion: {position}\n\tResultados precisos: {porc_preciso} %'.format(
#             param_reg_loss=param_reg_loss,
#             position=str(best_result["combinacion"]),
#             porc_preciso=round(best_result["precision"], 2)))



# # BUSQUEDA MEJORES PARAMETROS METRICOS
# for idx, metric in enumerate(metrics):
#     start_time = time.time()
#
#     model = Sequential()
#     model.add(Dense(16, input_dim=5, activation='elu'))
#     model.add(Dense(1, activation='relu'))
#
#     model.compile(loss='huber_loss',
#                   optimizer='adam',
#                   metrics=metric)
#     epochs = 100
#     best_result = entrenamiento_parametros(model, epochs)
#
#     end_time = time.time()
#     calc_time = end_time - start_time
#
#     s2hms((len(metrics) - (index + 1)) * calc_time)
#
#     combinaciones_parametros.append([metric, best_result["combinacion"], round(best_result["precision"], 2)])
#     print('Combinación más precisa:\n\tMetric: {metric}\n\tCombinacion: {position}\n\tResultados precisos: {porc_preciso} %'.format(
#             metric=metric,
#             position=str(best_result["combinacion"]),
#             porc_preciso=round(best_result["precision"], 2)))

# ------------------------------------------------------------

# combinaciones_parametros.sort(reverse=True, key=order_list)
#
# for combination in combinaciones_parametros:
#     print(combination)

# ------------------------------------------------------------

# Entrenamiento y guardado de mejor resultado diferente de 100
model = Sequential()
model.add(Dense(256, input_dim=5, activation='relu'))
model.add(Dense(128, activation='relu'))
model.add(Dropout(0.4))
model.add(Dense(5, activation='softmax'))

model.compile(loss='sparse_categorical_crossentropy',
              optimizer='adam',
              metrics=['accuracy'])
epochs = 20000
test_loss, test_acc = entrenamiento_parametros(model, epochs)

print('Combinación más precisa:\n\ttest_loss: {test_loss} %\n\ttest_acc: {test_acc} %'.format(test_loss=test_loss*100, test_acc=round(test_acc*100, 2)))
# print('Combinación más valida:\n\tCombinacion: {position}\n\tResultados validos: {porc_validos} %\n\tResultados precisos: {porc_preciso} %'.format(position = str(combinacion_mas_valida["combinacion"]), porc_validos=round(combinacion_mas_valida["validez"], 2), porc_preciso=round(combinacion_mas_valida["precision"], 2)))



# Cargamos modelo
# keras_model = keras.models.load_model('recursos/result_model_validez_keras.h5', compile = True)
# keras_model = keras.models.load_model('recursos/result_model_precision3_keras.h5', compile = True)
keras_model = keras.models.load_model('recursos/result_model_NN_keras.h5', compile = True)

aciertos_validos = []
aciertos_precisos = []
data = np.zeros((5,5), dtype=int)

for ind, linea in enumerate(testX):
    resp = keras_model.predict(np.array([linea]), verbose=0)

    resp = np.argmax(resp[0])

    data[int(np.round(resp))][int(testY[ind])] += 1

    if int(np.round(resp)) - 1 <= testY[ind] <= int(np.round(resp)) + 1:
        aciertos_validos.append(True)
    else:
        aciertos_validos.append(False)

    if testY[ind] == int(np.round(resp)):
        aciertos_precisos.append(True)
    else:
        aciertos_precisos.append(False)

porc_validos = round(100 * aciertos_validos.count(True) / len(aciertos_validos), 2)
porc_preciso = round(100 * aciertos_precisos.count(True) / len(aciertos_precisos), 2)

print('Resultados validos: {porc_validos}\nResultados precisos: {porc_preciso}'.format(porc_validos=str(porc_validos), porc_preciso=str(porc_preciso)))

results = keras_model.evaluate(testX, testY)
print("test loss, test acc:", results)

heat_map = sns.heatmap(data, annot=True)
plt.show()