# Autor: Mario Castañeda González
# TFG: Evaluación en línea del grado de involucración del usuario en actividades de interacción humano-robot

import cv2
import numpy as np
import os

# Función imprimir texto orientación cabeza
def head_text_view(rot_x, rot_y, z, image, location_ids):
    if rot_x < -10:
        text = "Cabeza a la izquierda"
    elif rot_x > 10:
        text = "Cabeza a la derecha"
    elif rot_y < -10:
        text = "Cabeza hacia abajo"
    elif rot_y > 10:
        text = "Cabeza hacia arriba"
    else:
        text = "Cabeza al frente"

    x = location_ids[0]
    y = location_ids[1]

    cv2.putText(image, text, (x + 10, y + 15), cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 0, 255), 1, cv2.LINE_AA)

# Función imprimir texto valores orientación cabeza
def head_text_pose(rot_x, rot_y, rot_z, image, location_ids):
    x = location_ids[0]
    h = location_ids[3]

    cv2.putText(image, "Cabeza (x, y, z): (" + str(np.round(rot_x,2)) + ", "
                + str(np.round(rot_y,2)) + ", " + str(np.round(rot_z,2)) + ")",
                (x + 10, h - 30), cv2.FONT_HERSHEY_SIMPLEX, 0.3, (0, 0, 255), 1)

# Función imprimir texto valores orientación cuerpo
def body_text_pose(rot_Y, image, location_ids):
    x = location_ids[0]
    h = location_ids[3]

    cv2.putText(image, "Torso (rot Y): (" + str(np.round(rot_Y,2)) + ")",
                (x + 10, h - 15), cv2.FONT_HERSHEY_SIMPLEX, 0.3, (0, 0, 255), 1)

# Función imprimir identidad persona
def show_id_text(image, location_ids, id_face):
    left = location_ids[0]
    top = location_ids[1]
    right = location_ids[2]
    bottom = location_ids[3]

    # Dibujar rectángulo en la cara
    cv2.rectangle(image, (left, top), (right, bottom), (0, 255, 0), 2)

    # Dibujar una etiqueta con el id de la cara
    cv2.rectangle(image, (left - 1, bottom), (right + 1, bottom + 30), (0, 255, 0), cv2.FILLED)
    font = cv2.FONT_HERSHEY_DUPLEX
    cv2.putText(image, id_face, (left + 6, bottom + 20), font, 0.8, (255, 255, 255), 1)

def show_id_text2(method, result, coord, image):
    x = coord[0]
    y = coord[1]
    w = coord[2]
    h = coord[3]

    dataPath = 'C:/Users/mario/PycharmProjects/HRIengagement/facial_recognition/data'
    imagePaths = os.listdir(dataPath)

    # EigenFaces
    if method == 'EigenFaces':
        if result[1] < 5700:
            cv2.putText(image, '{}'.format(imagePaths[result[0]]), (x, y - 25), 2, 0.7, (0, 255, 0), 1, cv2.LINE_AA)
            cv2.rectangle(image, (x, y), (w, h), (0, 255, 0), 2)
        else:
            cv2.putText(image, 'Desconocido', (x, y - 20), 2, 0.7, (0, 0, 255), 1, cv2.LINE_AA)
            cv2.rectangle(image, (x, y), (w, h), (0, 0, 255), 2)

    # FisherFace
    if method == 'FisherFaces':
        if result[1] < 500:
            cv2.putText(image, '{}'.format(imagePaths[result[0]]), (x, y - 25), 2, 0.7, (0, 255, 0), 1, cv2.LINE_AA)
            cv2.rectangle(image, (x, y), (w, h), (0, 255, 0), 2)
        else:
            cv2.putText(image, 'Desconocido', (x, y - 20), 2, 0.7, (0, 0, 255), 1, cv2.LINE_AA)
            cv2.rectangle(image, (x, y), (w, h), (0, 0, 255), 2)

    # LBPHFace
    if method == 'LBPH':
        if result[1] < 95:
            cv2.putText(image, '{}'.format(imagePaths[result[0]]), (x, y - 25), 2, 0.7, (0, 255, 0), 1, cv2.LINE_AA)
            cv2.rectangle(image, (x, y), (w, h), (0, 255, 0), 2)
        else:
            cv2.putText(image, 'Desconocido', (x, y - 20), 2, 0.7, (0, 0, 255), 1, cv2.LINE_AA)
            cv2.rectangle(image, (x, y), (w, h), (0, 0, 255), 2)

# Función imprimir texto emoción estimada
def show_emotion_text(result, coord, image):
    x = coord[0]
    y = coord[1]

    cv2.putText(image, '{}'.format(result), (x, y - 10), 2, 0.8, (0, 255, 0), 1, cv2.LINE_AA)


def show_engagement_text(emotion_max_frec, coord, image):
    y = coord[1]
    w = coord[2]

    cv2.putText(image, emotion_max_frec, (w - 70, y - 10), 2, 0.8, (0, 255, 0), 1, cv2.LINE_AA)

def print_FPS(fps, image):
    cv2.putText(image, 'FPS: {}'.format(round(fps, 2)), (20, 20), 2, 0.8, (0, 255, 0), 1, cv2.LINE_AA)